import os
import sys

sys.path.append(os.path.dirname(os.path.realpath(__file__)))


def getVersion():
    dir = os.path.dirname(os.path.realpath(__file__))
    version_file = os.path.join(dir, 'VERSION')
    with open(version_file, 'r') as f:
        version = f.read().strip()
    return version



#!/usr/bin/env python3

# This functionality should be merged into
# https://github.com/galaxyproject/training-material/blob/main/bin/check_instance.py

"""
This is a command line tool for checking whether a Galaxy workflow can be run on
a particular Galaxy instance. It does this by checking whether the tools used in
the workflow are available on the specified Galaxy instance. If a tool is not
available, the program can optionally install it from the tool shed.

Usage: check_workflow_tools.py [-h] [-i] workflow_path galaxy_url api_key

Optional arguments:
  -p PATH, --path PATH  path or URL to the workflow file
  -r REPO, --repo REPO  path to a local copy of the usegalaxy-tools repository
  -u URL, --url URL     Galaxy server URL
  -k KEY, --key KEY     an admin API key for the Galaxy server. Required to install the workflow and tools.
  -i, --install         install the workflow and any missing tools.
  -h, --help            show this help message and exit

"""

import os
import sys
import json
import yaml
import requests
import argparse

import bioblend.galaxy
from planemo.runnable import for_path
from planemo.galaxy.workflows import install_shed_repos


def check_workflow_tools(tools, workflow_path, galaxy_url, api_key, install_tools):
    """
    Check whether the tools used in a Galaxy workflow are available on a
    particular Galaxy instance. Optionally install missing tools from the
    tool shed.

    Parameters:
        tools (list): A list of tools to check.
        workflow_path (str): The path to the workflow file. Can be a local file path or a URL.
        galaxy_url (str): The URL of the Galaxy instance to check.
        api_key (str): An API key for accessing the Galaxy instance.
        install_tools (bool): A boolean indicating whether to install missing tools.

    Returns:
        int: An exit code of 0 if all tools are available, or 2 if any tools are missing or have the wrong version.
    """

    if workflow_path.startswith('http'):
        response = requests.get(workflow_path)
        if response.status_code != 200:
            print(f"Unable to fetch the workflow from {workflow_path}")
            print(f"{response.status_code} - {response.reason}")
            return
        workflow = response.text
    else:
        with open(workflow_path) as f:
            workflow = json.load(f)

    # print(json.dumps(workflow, indent=4))
    not_found = {}
    wrong_version = {}
    gi = bioblend.galaxy.GalaxyInstance(galaxy_url, api_key)
    nSteps = len(workflow['steps'])
    for nStep in range(nSteps):
        step = workflow['steps'][str(nStep)]
        if step['tool_id']:
            tool_id = step['tool_id']
            if 'tool_shed_repository' in step:
                revision = step['tool_shed_repository']['changeset_revision']
            else:
                # print(f"{tool_id} does not have a tool_shed_repository")
                # print(json.dumps(step, indent=4))
                revision = 'unknown'
            try:
                tool = gi.tools.show_tool(tool_id)
                # print(json.dumps(tool, indent=4))
                if tool['id'] != tool_id:
                    wrong_version[tool_id] = revision #tool['id']
            except:
                not_found[tool_id] = revision

    exit_code = 0
    if len(not_found) == 0 and len(wrong_version) == 0:
        print('All tools are available for this workflow')
    else:
        exit_code = 2
        process_not_found(not_found)
        process_wrong_version(wrong_version)

    if install_tools:
        if api_key is None:
            print("Unable to install tools without an admin API key.")
        else:
            result = gi.workflows.import_workflow_from_local_path(workflow_path, publish=True)
            print(json.dumps(result, indent=4))
            runnable = for_path(workflow_path)
            print("Installing tools")
            install_shed_repos(runnable, gi, False)

    return exit_code


def process_not_found(tools):
    """
    Process a list of missing tools and print out instructions for adding them
    to the appropriate lock files.

    Parameters:
        tools (dict): A dictionary of missing tools, where the keys are tool IDs and the values are tool revisions.
    """

    if len(tools) == 0:
        return

    print("# Please add the following entries to the appropriate .lock file(s) and update the matching .yml files.")
    for id, revision in tools.items():
        (shed, ignored, owner, group, name, version) = id.split('/')
        key = f"{owner}/{name}"
        if key in tools:
            if not revision in tools[key]['revisions']:
                print(f"We need to update {key} to revision {revision}")
            else:
                print(f"ERROR?: usegalaxy-tools/cloud already contains {key} revision {revision}!!!")
        else:
            section = {
                'name': name,
                'owner': owner,
                'revisions': [revision],
                'tool_panel_section_id': 'section_id',
                'tool_panel_section_label': 'section_label'
            }
            print(yaml.dump([section]))


def process_wrong_version(tools):
    """

    :param tools:
    :return:
    """
    if len(tools) == 0:
        return
    print("Wrong versions installed")
    for id, revision in tools.items():
        (shed, ignored, owner, group, name, version) = id.split('/')
        key = f"{owner}/{name}"
        if key in tools:
            tool = tools[key]
            if not revision in tool['spec']['revisions']:
                tool['spec']['revisions'].append(revision)
                # TODO if --update was specified on the command line then update the file.
                print(f"# Update the following block in  {tool['path']}")
                print(yaml.dump([tools[key]['spec']]))
            else:
                # If {key}@{revision} is not installed on the instance but is present
                # in usegalaxy-tools then something is amiss...
                print(f"usegalaxy-tools/cloud already contains {key} revision {revision} ")
        else:
            print("The wrong version is installed on Galaxy but there is no entry in usegalaxy-tools!")
            section = {
                'name': name,
                'owner': owner,
                'revisions': [revision],
                'tool_panel_section_id': 'section_id',
                'tool_panel_section_label': 'section_label'
            }
            print(yaml.dump([section]))


def load_lock_file(filepath, tools):
    """

    :param filepath:
    :param tools:
    :return:
    """
    # print(f"Loading {filepath}")
    with open(filepath) as f:
        tool_info = yaml.safe_load(f)

    for tool in tool_info['tools']:
        name = tool['name']
        owner = tool['owner']
        key = f"{owner}/{name}"
        if key in tools:
            # raise Exception(f"{tool} has already been defined in")
            print(f"WARNING: {key} is already defined in {tools[key]['path']}")
        else:
            tools[key] = {
                'path': filepath,
                'spec': tool
            }
        # revisions = tool['revisions']
        # if name not in tools:
        #     tools[name] = []
        # tools[name].extend(revisions)


def scan_usegalaxy_tools(directory_path):
    tools = {}
    files = os.scandir(directory_path)
    for item in files:
        if item.is_file and item.name.endswith('.lock'):
            load_lock_file(item.path, tools)
    return tools


def test_tool_not_found(id, url, key):
    print(f"Searching for {id}")
    gi = bioblend.galaxy.GalaxyInstance(url, key)
    print(json.dumps(gi.tools.show_tool(id), indent=4))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Check if all tools used in a workflow are present on a given Galaxy instance.',
        epilog='Copyright 2022 The Galaxy Project'
    )
    parser.add_argument('-p', '--path', required=True, help='path to the workflow file')
    parser.add_argument('-r', '--repo', required=True, help='path to a local copy of the usegalaxy-tools repository')
    parser.add_argument('-u', '--url', required=True, help='Galaxy server URL')
    parser.add_argument('-k', '--key', help='an admin API key for the Galaxy server.  Required to install the workflow and tools.')
    parser.add_argument('-i', '--install', default=False, action='store_true', help='install the workflow and any missing tools.')

    argv = parser.parse_args()
    exit_code = 0
    tools = scan_usegalaxy_tools(argv.repo)
    exit_code = check_workflow_tools(tools, argv.path, argv.url, argv.key, argv.install)
    sys.exit(exit_code)



from pprint import pprint

from .common import Context, connect, datasets


def list(context: Context, args: list):
    gi = connect(context)
    if len(args) == 0:
        for library in gi.libraries.get_libraries():
            print(f"{library['id']}\t{library['name']}\t{library['description']}")
        return
    folders = gi.libraries.show_library(args[0], contents=True)
    for folder in folders:
        print(f"{folder['id']}\t{folder['type']}\t{folder['name']}")


def create(context: Context, args: list):
    if len(args) != 2:
        print("ERROR: Invalid parameters, at least the name must be specified")
        return
    name = args[0]
    description = None
    if len(args) == 2:
        description = args[1]
    gi = connect(context)
    result = gi.libraries.create_library(name, description=description)
    pprint(result)


def delete(context: Context, args: list):
    print("library delete not implemented")


def upload(context: Context, args: list):
    if len(args) != 3:
        print(
            "ERROR: Invalid parameters. Specify the library and folder names and the dataset to upload"
        )
        return
    gi = connect(context)
    libraries = gi.libraries.get_libraries(name=args[0])
    if len(libraries) == 0:
        print(f"ERROR: No such library found: {args[0]}")
        return
    if len(libraries) > 1:
        print(f"WARNING: more than one library name matched. Using the first one found")
        for library in libraries:
            print(f"{library['id']}\t{library['create_time']}\t{library['name']}")

    library_id = libraries[0]['id']
    folders = gi.libraries.get_folders(library_id, name=args[1])
    if len(folders) == 0:
        print(f"ERROR: no folder with the name: {args[1]}")
        return
    folder_id = folders[0]['id']
    dataset_url = datasets['dna'][int(args[2])]
    result = gi.libraries.upload_file_from_url(
        library_id, dataset_url, folder_id=folder_id
    )
    pprint(result)
    return

    # library_id = args[0]
    # if len(args) == 3:
    #     folder_id = args[1]
    #     file_url = datasets['dna'][int(args[2])]
    # else:
    #     file_url = datasets['dna'][int(args[1])]
    # # ftp://ftp.sra.ebi.ac.uk/vol1/fastq/SRR592/SRR592109/SRR592109_2.fastq.gz
    # result = gi.libraries.upload_file_from_url(library_id, file_url, folder_id=folder_id)
    # pprint(result)


def download(context: Context, args: list):
    print("library download not implemented")


def show(context: Context, args: list):
    print("library show not implemented")



import json
import os
import sys

sys.path.append(os.path.dirname(os.path.realpath(__file__)))

# from common import parse_profile

INVOCATIONS_DIR = "invocations"
METRICS_DIR = "metrics"

parser = None


class Keys:
    NAME = 'name'
    RUNS = 'runs'
    INPUTS = 'inputs'
    REFERENCE_DATA = 'reference_data'
    WORKFLOW_ID = 'workflow_id'
    DATASET_ID = 'dataset_id'
    COLLECTION = 'collection'
    HISTORY_BASE_NAME = 'output_history_base_name'
    HISTORY_NAME = 'history_name'



import sys
import time
from concurrent.futures import ThreadPoolExecutor

def worker(delay: int):
    print("Starting worker")
    time.sleep(delay)
    print("Worker woke up")
    return "Done"


def run():
    with ThreadPoolExecutor(max_workers=4) as executor:
        print("Submitting job")
        f1 = executor.submit(worker, 30)
        try:
            print(f1.result(timeout=2))
        except:
            print("Caught a time out error")
            sys.exit()



if __name__ == '__main__':
    run()
    print("End of program")




import json
import logging
import time
from pprint import pprint

from .common import Context, connect, find_history, print_json

log = logging.getLogger('abm')


def list(context: Context, args: list):
    state = ''
    history_id = None
    log.debug('Processing args')
    log_state = False
    while len(args) > 0:
        arg = args.pop(0)
        if arg in ['-s', '--state', 'state']:
            if len(args) == 0:
                print("ERROR: specify a state, eg 'ok', 'error'")
                return
            state = args.pop(0)
            log_state = True
        elif arg in ['-h', '--history']:
            if len(args) == 0:
                print("ERROR: history ID was not specified.")
                return
            history_id = args.pop(0)
            log.debug(f"Getting jobs from history {history_id}")
    log.debug('Connecting to the Galaxy server')
    gi = connect(context)
    if log_state:
        log.debug(f"Getting jobs with state {state}")
    else:
        log.debug("Getting job list")
    if history_id:
        history_id = find_history(gi, history_id)
        if history_id is None:
            print("ERROR: No such history")
            return
    job_list = gi.jobs.get_jobs(state=state, history_id=history_id)
    log.debug(f"Iterating over job list with {len(job_list)} items")
    for job in job_list:
        print(f"{job['id']}\t{job['state']}\t{job['update_time']}\t{job['tool_id']}")


def show(context: Context, args: list):
    if len(args) != 1:
        print("ERROR: Invalid parameters. Job ID is required")
        return
    gi = connect(context)
    job = gi.jobs.show_job(args[0], full_details=True)
    print(json.dumps(job, indent=4))


def wait(context: Context, args: list):
    if len(args) != 1:
        print("ERROR: Invalid parameters. Job ID is required")
        return
    gi = connect(context)
    state = "Unknown"
    waiting = True
    while waiting:
        job = gi.jobs.show_job(args[0], full_details=False)
        state = job["state"]
        if state == "ok" or state == "error":
            waiting = False
        else:
            time.sleep(15)
    print(json.dumps(job, indent=4))


def get_value(metric: dict):
    if metric['name'] == 'runtime_seconds':
        return metric['raw_value']
    return metric['value']


def metrics(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: no job ID provided")
        return
    gi = connect(context)
    if len(args) > 1:
        arg = args.pop(0)
        if arg in ['-h', '--history']:
            history_id = args.pop(0)
            log.debug(f"Getting metrics for jobs from history {history_id}")
            job_list = gi.jobs.get_jobs(history_id=history_id)
            metrics = []
            for job in job_list:
                metrics.append(
                    {
                        'job_id': job['id'],
                        'job_state': job['state'],
                        'tool_id': job['tool_id'],
                        'job_metrics': gi.jobs.get_metrics(job['id']),
                    }
                )
        else:
            print(f"ERROR: Unrecognized argument {arg}")
    else:
        job = gi.jobs.show_job(args[0])
        metrics = [
            {
                'job_id': job['id'],
                'job_state': job['state'],
                'tool_id': job['tool_id'],
                'job_metrics': gi.jobs.get_metrics(args[0]),
            }
        ]
    print(json.dumps(metrics, indent=4))
    # metrics = {}
    # for m in gi.jobs.get_metrics(args[0]):
    #     metrics[m['name']] = get_value(m)
    # try:
    #     print(f"{metrics['galaxy_slots']},{metrics['galaxy_memory_mb']},{metrics['runtime_seconds']}")
    # except:
    #     print(',,')


def cancel(context: Context, args: list):
    if len(args) == 0:
        print('ERROR: no job ID provided.')
        return
    gi = connect(context)
    state = ''
    history = None
    jobs = []
    while len(args) > 0:
        arg = args.pop(0)
        if arg in ['-s', '--state']:
            state = args.pop(0)
        elif arg in ['-h', '--history']:
            history = find_history(gi, args.pop(0))
            if history is None:
                print("ERROR: No such history")
                return
    if state or history:
        if len(jobs) > 0:
            print(
                "ERROR: To many parameters. Either filter by state or history, or list job IDs"
            )
            return
        jobs = [job['id'] for job in gi.jobs.get_jobs(state=state, history_id=history)]
    for job in jobs:
        if gi.jobs.cancel_job(job):
            print(f"Job {job} canceled")
        else:
            print(
                f"ERROR: Unable to cancel {job}, job was already in a terminal state."
            )


def problems(context: Context, args: list):
    if len(args) == 0:
        print('ERROR: no job ID provided.')
        return
    gi = connect(context)
    print_json(gi.jobs.get_common_problems(args[0]))


def rerun(context: Context, args: list):
    remap = False
    if '-r' in args:
        remap = True
        args.remove('-r')
    if '--remap' in args:
        remap = True
        args.remove('--remap')
    if len(args) == 0:
        print("ERROR: no job ID provided")
        return
    gi = connect(context)
    result = gi.jobs.rerun_job(args[0], remap=remap)
    print_json(result)



import configparser
import json
import os
import traceback

import arrow
import requests
from cloudlaunch_cli.main import create_api_client
from common import Context

BOLD = '\033[1m'
CLEAR = '\033[0m'


def h1(text):
    return f"{BOLD}{text}{CLEAR}"


list_help = f'''
{h1('NAME')}
    list
    
{h1('DESCRIPTION')}
    Lists deployments
    
{h1('SYNOPSIS')}
    abm cloudlaunch list [OPTIONS]
    
{h1('OPTIONS')}
    -a, --archived  list deployments that have been archived
    -r, --running   list running deployments
    -d, --deleted   list deleted deployments
    -l, --launch    list deploymentes that are launching or have launched
    -n, --limit     limit output to N lines
    -h, --help      print this help message
    
{h1('NOTES')}
    The --archived, --running, and --deleted options are mutually exclusive.
    
'''


def list(context: Context, args: list):
    archived = False
    filter = None
    status = lambda t: t.instance_status if t.instance_status else t.status
    n = None
    while len(args) > 0:
        arg = args.pop(0)
        if arg in ['-a', '--archived', 'archived']:
            archived = True
        elif arg in ['-r', '--running', 'running']:
            filter = lambda d: 'running' in status(d.latest_task) or (
                'LAUNCH' == d.latest_task.action and 'SUCCESS' == status(d.latest_task)
            )
        elif arg in ['-d', '--deleted', 'deleted']:
            filter = lambda d: 'DELETE' in d.latest_task.action
        elif arg in ['-l', '--launch', 'launch']:
            filter = lambda d: 'LAUNCH' in d.latest_task.action
        elif arg in ['-n', '--limit', 'limit']:
            n = int(args.pop(0))
        elif arg in ['-h', '--help', 'help']:
            print(list_help)
            return
        else:
            print(f"Invalid parameter: {arg}")
            return
    deployments = create_api_client().deployments.list(archived=archived)

    if filter is not None:
        deployments = [d for d in deployments if filter(d)]

    if n is not None and len(deployments) > n:
        deployments = deployments[:n]
    _print_deployments(deployments)


def create(context: Context, args: list):
    cloud = None
    region = None
    params = {'application': 'cloudman-20', 'application_version': 'dev'}
    config = {
        "config_cloudlaunch": {
            "rootStorageType": "volume",
            "rootStorageSize": 42,
            "keyPair": "",
        },
        "config_cloudman2": {"clusterPassword": "gvl_letmein"},
    }
    targets = {'aws': 11, 'gcp': 16}
    regions = {
        'aws': {
            'us-east-1': 11,
            'us-east-2': 12,
            'us-west-1': 13,
            'us-west-2': 14,
            'us-east-1b': 36,
        },
        'gcp': {'us-central1': 16},
    }
    while len(args) > 0:
        arg = args.pop(0)
        if arg in ['aws', 'gcp']:
            if cloud is not None:
                print(f"ERROR: the cloud provider has already been specified: {cloud}")
                return
            cloud = arg
            # params['deployment_target_id'] = targets[cloud]
        elif arg in ['-c', '--config']:
            filepath = args.pop(0)
            with open(filepath, 'r') as f:
                params['config_app'] = json.load(f)
        elif arg in ['-t', '--type']:
            config['config_cloudlaunch']['vmType'] = args.pop(0)
        elif arg in ['-k', '--kp', '--keypair']:
            config['config_cloudlaunch']['keyPair'] = args.pop(0)
        elif arg in ['-p', '--password']:
            config['config_cloudman2']['clusterPassword'] = args.pop(0)
        elif arg in ['-r', '--region']:
            region = args.pop(0)
        elif 'name' in params:
            print(
                f"ERROR: the cluster name has already been specified: {params['name']}"
            )
            return
        else:
            params['name'] = arg

    params['config_app'] = config
    if 'name' not in params:
        print("ERROR: cluster name not specifed")
        return
    if cloud is None:
        print("ERROR: cloud provider not specied. Must be one of 'aws' or 'gcp'")
        return
    if region is not None:
        if cloud not in regions:
            print("ERROR: No regions have been defined for cloud provider {cloud")
            return
        if region not in regions[cloud]:
            print(f"ERROR: Region {region} has not been defined for {cloud}")
            return
        params['deployment_target_id'] = regions[cloud][region]

    # If the deployment target (region) has not been specified on the command
    # line use the default for that provider.
    if 'deployment_target_id' not in params:
        params['deployment_target_id'] = targets[cloud]

    if 'vmType' not in config['config_cloudlaunch']:
        print("ERROR: please specify a VM type.")
        return
    print(f"Deployment target id {params['deployment_target_id']}")
    cloudlaunch_client = create_api_client(cloud)
    try:
        new_deployment = cloudlaunch_client.deployments.create(**params)
        _print_deployments([new_deployment])
    except Exception as e:
        print("Unable to launch the cluster")
        # traceback.print_tb(e.__traceback__)
        print(e)


def delete(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: Invalid parameters.")
        return
    # if args[0] not in ['aws', 'gcp']:
    #     print(f"ERROR: Invalid cloud specified: '{args[0]}'. Must be one of 'aws' or 'gcp'.")
    #     return
    # id = args[0]
    configfile = os.path.expanduser("~/.cloudlaunch")
    if not os.path.exists(configfile):
        print("ERROR: Cloudlaunch has not been configured.")
        return

    config = configparser.ConfigParser()
    config.read(configfile)

    # cloudlaunch_client = create_api_client(args[0])
    # cloudlaunch_client.deployments.delete(args[1])
    url = config['cloudlaunch-cli']['url']
    token = config['cloudlaunch-cli']['token']
    headers = {
        'Accept': 'application/json',
        'Content-type': 'application/json',
        'Authorization': f"Token {token}",
    }

    data = dict(action='DELETE')
    for id in args:
        print(f"URL is: {url}/deployments/{id}/tasks/")
        response = requests.post(
            f"{url}/deployments/{id}/tasks/", json=data, headers=headers
        )
        if response.status_code < 300:
            print(f"Deleted deployment {id}")
        else:
            print(f"{response.status_code} - {response.reason}")
            print(response.text)
        print()


def _print_deployments(deployments):
    if len(deployments) > 0:
        print(
            "{:6s}  {:24s}  {:6s}  {:15s}  {:15s} {:s}".format(
                "ID", "Name", "Cloud", "Created", "Address", "Status"
            )
        )
    else:
        print("No deployments.")
    for deployment in deployments:
        created_date = arrow.get(deployment.added)
        latest_task = deployment.latest_task
        latest_task_status = (
            latest_task.instance_status
            if latest_task.instance_status
            else latest_task.status
        )
        latest_task_display = "{action}:{latest_task_status}".format(
            action=latest_task.action, latest_task_status=latest_task_status
        )
        ip_address = deployment.public_ip if deployment.public_ip else 'N/A'
        cloud = deployment._data['deployment_target']['target_zone']['cloud']['id']
        print(
            "{identifier:6d}  {name:24.24s}  {cloud:6.6s}  {created_date:15.15s}  "
            "{ip_address:15.15s} {latest_task_display}".format(
                identifier=deployment._id,
                cloud=cloud,
                created_date=created_date.humanize(),
                latest_task_display=latest_task_display,
                ip_address=ip_address,
                **deployment._data,
            )
        )
        # pprint(deployment._data)
        # print()



import json
import os
from pathlib import Path
from pprint import pprint

import yaml
from bioblend.galaxy import dataset_collections
from common import (Context, _get_dataset_data, _make_dataset_element, connect,
                    find_history, print_json)


def list(context: Context, args: list):
    gi = connect(context)
    kwargs = {'limit': 10000, 'offset': 0}
    if len(args) > 0:
        if args[0] in ['-s', '--state']:
            if len(args) != 2:
                print("ERROR: Invalid command.")
                return
            kwargs['state'] = args[1]
        else:
            print(f"ERROR: Invalid parameter: {args[0]}")
            return
    # datasets = gi.datasets.get_datasets(limit=10000, offset=0)  # , deleted=True, purged=True)
    datasets = gi.datasets.get_datasets(**kwargs)
    if len(datasets) == 0:
        print('No datasets found')
        return
    print(f'Found {len(datasets)} datasets')
    print('ID\tHistory\tType\tDeleted\tState\tName')
    for dataset in datasets:
        state = dataset['state'] if 'state' in dataset else 'unknown'
        print(
            f"{dataset['id']}\t{dataset['history_id']}\t{dataset['history_content_type']}\t{dataset['deleted']}\t{state}\t{dataset['name']}"
        )


def clean(context: Context, args: list):
    if len(args) == 0:
        invalid_states = ['error', 'discarded', 'unknown']
    else:
        invalid_states = args
    gi = connect(context)
    datasets = gi.datasets.get_datasets(
        limit=10000, offset=0
    )  # , deleted=True, purged=True)
    if len(datasets) == 0:
        print('No datasets found')
        return
    for dataset in datasets:
        state = dataset['state'] if 'state' in dataset else 'unknown'
        if not dataset['deleted'] and state in invalid_states:
            gi.histories.delete_dataset(dataset['history_id'], dataset['id'], True)
            print(f"Removed {dataset['id']}\t{state}\t{dataset['name']}")


def show(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: no dataset ID provided")
        return
    gi = connect(context)
    result = gi.datasets.show_dataset(args[0])
    print(json.dumps(result, indent=4))


def get(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: no dataset ID provided")
        return
    gi = connect(context)
    result = gi.datasets.get_datasets(args[0])
    print(json.dumps(result, indent=4))


def delete(context: Context, args: list):
    # gi = connect(context)
    print("dataset delete not implemented")


def upload(context: Context, args: list):
    history = None
    name = None
    url = None
    gi = None
    while len(args) > 0:
        arg = args.pop(0)
        if arg in ['--hs', '--hist', '--history']:
            history = args.pop(0)
        elif arg in ['-c', '--create']:
            gi = connect(context)
            history = gi.histories.create_history(args.pop(0)).get('id')
        elif arg in ['-n', '--name']:
            name = args.pop(0)
        elif url is not None:
            print(f'ERROR: invalid option. URL already set to {arg}')
            return
        else:
            url = arg
    if history is None:
        print("ERROR: please specify or create a history.")
        return
    if gi is None:
        gi = connect(context)
    if name:
        _import_from_url(gi, history, url, file_name=name)
    else:
        _import_from_url(gi, history, url)


def collection(context: Context, args: list):
    type = 'list:paired'
    collection_name = 'collection'
    elements = []
    hid = None
    gi = connect(context)
    while len(args) > 0:
        arg = args.pop(0)
        if arg == '-t' or arg == '--type':
            type = args.pop(0)
        elif arg == '-n' or arg == '--name':
            collection_name = args.pop(0)
        elif '=' in arg:
            name, value = arg.split('=')
            dataset = _get_dataset_data(gi, value)
            if dataset is None:
                print(f"ERROR: dataset not found {value}")
                return
            # print_json(dataset)
            if hid is None:
                hid = dataset['history']
            elif hid != dataset['history']:
                print('ERROR: Datasets must be in the same history')
            elements.append(_make_dataset_element(name, dataset['id']))

    if len(elements) == 0:
        print("ERROR: No dataset elements have been defined for the collection")
        return
    result = gi.histories.create_dataset_collection(
        history_id=hid,
        collection_description=dataset_collections.CollectionDescription(
            name=collection_name, type=type, elements=elements
        ),
    )
    print(json.dumps(result, indent=4))


def import_from_config(context: Context, args: list):
    gi = None
    key = None
    history = None
    kwargs = {}
    while len(args) > 0:
        arg = args.pop(0)
        if arg in ['--hs', '--hist', '--history']:
            history = args.pop(0)
        elif arg in ['-c', '--create']:
            gi = connect(context)
            history = gi.histories.create_history(args.pop(0)).get('id')
        elif arg in ['-n', '--name']:
            kwargs['file_name'] = args.pop(0)
        elif key is not None:
            print(f"ERROR: key already set: {key}")
            return
        else:
            key = arg

    configfile = os.path.join(Path.home(), '.abm', 'datasets.yml')
    if not os.path.exists(configfile):
        print("ERROR: ABM has not been configured to import datasets.")
        print(f"Please create {configfile}")
        return

    with open(configfile, 'r') as f:
        datasets = yaml.safe_load(f)
    if not key in datasets:
        print(f"ERROR: dataset {key} has not been defined.")
        return
    url = datasets[key]

    if gi is None:
        gi = connect(context)
    if history is not None:
        history = find_history(gi, history)

    response = gi.tools.put_url(url, history, **kwargs)
    print(json.dumps(response, indent=4))


def _import_from_url(gi, history, url, **kwargs):
    response = gi.tools.put_url(url, history, **kwargs)
    print(json.dumps(response, indent=4))


def download(context: Context, args: list):
    gi = connect(context)
    if len(args) == 0:
        print('ERROR: no dataset ID given')
        return
    elif len(args) > 1:
        pprint(gi.datasets.download_dataset(args[0], file_path=args[1]))
    else:
        pprint(gi.datasets.download_dataset(args[0]))


def find(context: Context, args: list):
    if len(args) == 0:
        print('ERROR: no dataset name given.')
        return
    gi = connect(context)
    datasets = gi.datasets.get_datasets(name=args[0])
    if len(datasets) == 0:
        print('WARNING: no datasets found with that name')
        return
    if len(datasets) > 1:
        print(f'WARNING: found {len(datasets)} datasets with that name.')

    for ds in datasets:
        pprint(ds)


def rename(context: Context, args: list):
    if len(args) != 3:
        print("ERROR: please provide the history ID, dataset ID, and new name.")
        return
    gi = connect(context)
    response = gi.histories.update_dataset(args[0], args[1], name=args[2])
    result = {'state': response['state'], 'name': response['name']}
    print(json.dumps(result, indent=4))


def test(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: no dataset ID provided")
        return
    gi = connect(context)
    dataset = gi.datasets.show_dataset(args[0])
    if dataset is None:
        print(f"ERROR: no dataset with ID {args[0]}")
        return
    name = dataset['name']
    print(f"Found dataset with name {name}")
    print(yaml.dump(dataset))
    datasets = gi.datasets.get_datasets(name=name)
    if datasets is None or len(datasets) == 0:
        print(f"ERROR: could not find a dataset named {name}")
        return
    print(f"Found {len(datasets)} datasets")
    for dataset in datasets:
        print(f"{dataset['id']}\t{dataset['name']}")



import json
import logging
import os
from pathlib import Path
from pprint import pprint

import planemo
import requests
import yaml
from common import Context, connect, summarize_metrics
from planemo.galaxy.workflows import install_shed_repos
from planemo.runnable import for_path, for_uri

log = logging.getLogger('abm')


def list(context: Context, args: list):
    gi = connect(context)
    workflows = gi.workflows.get_workflows(published=True)
    if len(workflows) == 0:
        print('No workflows found')
        return
    print(f'Found {len(workflows)} workflows')
    for workflow in workflows:
        print(f"{workflow['id']}\t{workflow['name']}")


def delete(context: Context, args: list):
    if len(args) == 0:
        print(f'ERROR: no workflow ID given.')
        return
    gi = connect(context)
    print(gi.workflows.delete_workflow(args[0]))


def upload(context: Context, args: list):
    if len(args) == 0:
        print('ERROR: no workflow file given')
        return
    path = args[0]
    if path.startswith('http'):
        import_from_url(context, args)
        return
    if not os.path.exists(path):
        print(f'ERROR: file not found: {path}')
        return
    gi = connect(context)
    print("Importing the workflow")
    pprint(gi.workflows.import_workflow_from_local_path(path, publish=True))
    runnable = for_path(path)
    print("Installing tools")
    result = install_shed_repos(runnable, gi, False)
    pprint(result)


def import_from_url(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: no workflow URL given")
        return
    url = args[0]

    # There is a bug in ephemeris (for lack of a better term) that assumes all
    # Runnable objects can be found on the local file system
    input_text = None
    filename = url.split('/')[-1]
    cache = os.path.expanduser("~/.abm/cache/workflows")
    if not os.path.exists(cache):
        os.makedirs(cache)
    cached_file = os.path.join(cache, filename)
    if os.path.exists(cached_file):
        with open(cached_file) as f:
            input_text = f.read()
    else:
        response = requests.get(url)
        if response.status_code != 200:
            print(
                f"ERROR: There was a problem downloading the workflow: {response.status_code}"
            )
            print(response.reason)
            return
        input_text = response.text
        with open(cached_file, 'w') as f:
            f.write(input_text)

    # response = requests.get(url)
    # if (response.status_code != 200):
    #     print(f"ERROR: There was a problem downloading the workflow: {response.status_code}")
    #     print(response.reason)
    #     return
    try:
        workflow = json.loads(input_text)
    except Exception as e:
        print("ERROR: Unable to parse workflow")
        print(e)
        return

    gi = connect(context)
    result = gi.workflows.import_workflow_dict(workflow, publish=True)
    print(json.dumps(result, indent=4))
    runnable = for_path(cached_file)
    # runnable = for_uri(url)
    print("Installing tools")
    result = install_shed_repos(runnable, gi, False, install_tool_dependencies=True)
    pprint(result)


def import_from_config(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: no workflow ID given")
        return
    key = args[0]
    userfile = os.path.join(Path.home(), ".abm", "workflows.yml")
    if not os.path.exists(userfile):
        print("ERROR: this instance has not been configured to import workflows.")
        print(f"Please configure {userfile} to enable workflow imports")
        return
    with open(userfile, 'r') as f:
        workflows = yaml.safe_load(f)
    if not key in workflows:
        print(f"ERROR: no such workflow: {key}")
        return

    url = workflows[key]
    import_from_url(context, [url])


def download(context: Context, args: list):
    if len(args) == 0:
        print('ERROR: no workflow ID given')
        return
    gi = connect(context)
    workflow = json.dumps(gi.workflows.export_workflow_dict(args[0]), indent=4)
    if len(args) == 2:
        with open(args[1], 'w') as f:
            f.write(workflow)
            print(f'Wrote {args[1]}')
    else:
        print(workflow)


def show(context: Context, args: list):
    if len(args) == 0:
        print('ERROR: no workflow ID given')
        return
    gi = connect(context)
    result = gi.workflows.show_workflow(args[0])
    print(json.dumps(result, indent=4))


def inputs(context: Context, args: list):
    if len(args) == 0:
        print('ERROR: no workflow ID given')
        return
    gi = connect(context)
    workflow = gi.workflows.show_workflow(args[0])
    inputs = workflow['inputs']
    for input_name, input_dict in inputs.items():
        print(input_name)
        print(json.dumps(input_dict, indent=4))


def invocation(context: Context, args: list):
    if len(args) != 2:
        print("ERROR: Invalid paramaeters. A workflow ID invocation ID are required")
        return
    workflow_id = None
    invocation_id = None
    while len(args) > 0:
        arg = args.pop(0)
        if arg in ['-w', '--work', '--workflow']:
            print("Setting workflow id")
            workflow_id = args.pop(0)
        # elif arg in ['-i', '--invoke', '--invocation']:
        #     invocation_id = args.pop(0)
        #     print("Setting invocation id")
        else:
            print(f'Invalid parameter: "{arg}')
            return
    if workflow_id is None:
        print("ERROR: No workflow ID provided")
        return
    # if invocation_id is None:
    #     print("ERROR: No invocation ID provided")
    #     return
    gi = connect(context)
    # result = gi.workflows.show_invocation(workflow_id, invocation_id)
    invocations = gi.invocations.get_invocations(
        workflow_id=workflow_id, view='element', step_details=True
    )
    # print(json.dumps(result, indent=4))
    print('ID\tState\tWorkflow\tHistory')
    for invocation in invocations:
        id = invocation['id']
        state = invocation['state']
        workflow = invocation['workflow_id']
        history = invocation['history_id']
        print(f'{id}\t{state}\t{workflow}\t{history}')


def find(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: no workflow name given")
        return
    gi = connect(context)
    pprint(gi.workflows.get_workflows(name=args[0]))


def test(context: Context, args: list):
    # gi = connect(context)
    # print(f"Searching for workflow {args[0]}")
    # flows = gi.workflows.get_workflows(name=args[0], published=True)
    # pprint(flows)
    log.debug('debug')
    log.info('info')
    log.warning('warn')
    log.error('error')
    log.critical('critical')


def publish(context: Context, args: list):
    if len(args) != 1:
        print("USAGE: publish ID")
        return
    gi = connect(context)
    result = gi.workflows.update_workflow(args[0], published=True)
    print(f"Published: {result['published']}")


def rename(context: Context, args: list):
    if len(args) != 2:
        print("USAGE: rename ID 'new workflow name'")
        return
    gi = connect(context)
    result = gi.workflows.update_workflow(args[0], name=args[1])
    print(f"Renamed workflow to {result['name']}")


def summarize(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: Provide one or more workflow ID values.")
        return
    gi = connect(context)
    wid = args[0]
    all_jobs = []
    invocations = gi.invocations.get_invocations(workflow_id=wid)
    for invocation in invocations:
        id = invocation['id']
        jobs = gi.jobs.get_jobs(invocation_id=id, workflow_id=wid)
        for job in jobs:
            job['invocation_id'] = id
            job['workflow_id'] = wid
            all_jobs.append(job)
    summarize_metrics(gi, all_jobs)



import json
import os
import sys
import time
from pathlib import Path
from pprint import pprint

import yaml
from bioblend.galaxy.objects import GalaxyInstance
from lib.common import (Context, connect, find_history, parse_profile,
                        print_json, summarize_metrics)

#
# History related functions
#


def longest_name(histories: list):
    longest = 0
    for history in histories:
        if len(history['name']) > longest:
            longest = len(history['name'])
    return longest


def pad(value: bool):
    if value:
        return 'True   '
    return 'False  '


def print_histories(histories: list):
    if len(histories) == 0:
        print("There are no available histories.")
        return

    id_width = len(histories[0]['id'])
    name_width = longest_name(histories)

    print(f"{'ID':<{id_width}} {'Name':<{name_width}} Deleted Public  Tags")
    for history in histories:
        print(
            f"{history['id']:<{id_width}} {history['name']:<{name_width}} {pad(history['deleted'])} {pad(history['published'])} {', '.join(history['tags'])}"
        )


def _list(context: Context, args: list):
    gi = connect(context)
    print_histories(gi.histories.get_histories())

    if len(args) > 0:
        if args[0] in ['all', '-a', '--all']:
            print('Histories Published by all users')
            print_histories(gi.histories.get_published_histories())


def show(context: Context, args: list):
    contents = False
    if '-c' in args:
        contents = True
        args.remove('-c')
    if '--contents' in args:
        contents = True
        args.remove('--contents')
    if len(args) == 0:
        print("ERROR: No history ID provided.")
        return
    gi = connect(context)
    hid = find_history(gi, args[0])
    if hid is None:
        print(f"ERROR: No such history {args[0]}")
        return
    history = gi.histories.show_history(hid, contents=contents)
    print(json.dumps(history, indent=4))


def find(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: No history ID provided")
        return
    return_json = False
    history_name = None
    while len(args) > 0:
        arg = args.pop(0)
        if arg in ['-j', '--json']:
            return_json = True
        else:
            history_name = arg
    if history_name is None:
        print("ERROR: no history name provided")
        return
    gi = connect(context)
    histories = gi.histories.get_histories(name=history_name)
    if return_json:
        print_json(histories)
    else:
        for history in histories:
            print(f"{history['id']}\t{history['name']}")


def clean(context: Context, args: list):
    gi = connect(context)
    print("Empty histories")
    for history in gi.histories.get_histories():
        info = gi.histories.show_history(history['id'])
        if 'empty' in info:
            if info['empty']:
                print(f"{history['id']}\t{history['name']}")
            else:
                pprint(info)
        else:
            pprint(info)
            return


def download(context: Context, args: list):
    print('history download not implemented')


def upload(context: Context, args: list):
    print('history upload not implemented')


def test(context: Context, args: list):
    gi = connect(context)
    gi.datasets.publish_dataset('bbd44e69cb8906b5a5560acb9ce77faa', published=True)


def export(context: Context, args: list):
    wait = True
    if '--no-wait' in args:
        wait = False
        args.remove('--no-wait')
    if '-n' in args:
        wait = False
        args.remove('-w')
    if len(args) == 0:
        print("ERROR: no history ID specified")
        return
    gi = connect(context)
    hid = find_history(gi, args[0])
    jeha_id = gi.histories.export_history(hid, gzip=True, wait=wait)
    # global GALAXY_SERVER
    export_url = "unknown"
    if wait:
        export_url = f"{context.GALAXY_SERVER}/history/export_archive?id={args[0]}&jeha_id={jeha_id}"
        print(f"The history can be imported from {export_url}")
        history = gi.histories.show_history(hid, contents=False)
        tags = history['tags']
        if 'exported' not in tags:
            tags.append('exported')
            gi.histories.update_history(hid, tags=tags)
            print(f"History tagged with: {tags}")

    else:
        print("Please run the following command to obtain the ID of the export job:")
        print("python abm <cloud> job list | grep EXPORT")
        print()
    return export_url


def publish(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: No history ID provided.")
        return
    gi = connect(context)
    result = gi.histories.update_history(find_history(gi, args[0]), published=True)
    print(f"Published: {result['published']}")


def rename(context: Context, args: list):
    if len(args) != 2:
        print("ERROR: Invalid command")
        print("USAGE: rename ID 'new history name'")
        return
    gi = connect(context)
    result = gi.histories.update_history(find_history(gi, args[0]), name=args[1])
    print(f"History renamed to {result['name']}")


def _import(context: Context, args: list):
    gi = connect(context)
    result = gi.histories.import_history(url=args[0])
    id = result['id']
    try:
        gi.jobs.wait_for_job(id, 86400, 10, False)
    except:
        return False
    return True


def himport(context: Context, args: list):
    def error_message(msg='Invalid command'):
        print(f"ERROR: {msg}")
        print(f"USAGE: {sys.argv[0]} history import SERVER HISTORY_ID JEHA_ID")
        print(f"       {sys.argv[0]} history import http://GALAXY_SERVER_URL")
        print(f"       {sys.argv[0]} history import [dna|rna]")

    wait = True
    if '-n' in args:
        args.remove('-n')
        wait = False
    if '--no-wait' in args:
        args.remove('--no-wait')
        wait = False

    if len(args) == 1:
        if 'http' in args[0]:
            url = args[0]
        else:
            datasets = None
            config = f'{os.path.dirname(os.path.abspath(__file__))}/histories.yml'
            # First load the histories.yml file that is pacakged with abm
            if os.path.exists(config):
                with open(config, 'r') as f:
                    datasets = yaml.safe_load(f)
            # Then load the user histories.yml, if any
            userfile = os.path.join(Path.home(), ".abm", "histories.yml")
            if os.path.exists(userfile):
                if datasets is None:
                    datasets = {}
                with open(userfile, 'r') as f:
                    userdata = yaml.safe_load(f)
                    for key, item in userdata.items():
                        datasets[key] = item
            if datasets is None:
                error_message("No history URLs have been configured.")
                return
            if not args[0] in datasets:
                error_message('Please specify a URL or name of the history to import')
                return
            url = datasets[args[0]]
    elif len(args) == 3:
        server, key = parse_profile(args[0])
        if server is None:
            error_message(f"Invalid server profile name: {args[0]}")
            return
        url = f"{server}history/export_archive?id={args[1]}&jeha_id={args[2]}"
    else:
        error_message()
        return

    gi = connect(context)
    print(f"Importing history from {url}")
    result = gi.histories.import_history(url=url)
    if wait:
        id = result['id']
        print(f"Waiting for job {id}")
        try:
            gi.jobs.wait_for_job(id, 86400, 10, False)
            # TODO We could rename the history here if we wanted to.
            print('Done')
        except:
            return False
    else:
        print(json.dumps(result, indent=4))
    return True


def create(context: Context, args: list):
    if len(args) != 1:
        print("ERROR: Please provide a history name")
        return
    gi = connect(context)
    id = gi.histories.create_history(args[0])
    print(json.dumps(id, indent=4))


def delete(context: Context, args: list):
    if len(args) != 1:
        print('ERROR: please provide the history ID')
        return
    gi = connect(context)
    history = find_history(gi, args[0])
    if history is None:
        print("ERROR: No such history.")
    gi.histories.delete_history(history, True)
    print(f"Deleted history {args[0]}")


def copy(context: Context, args: list):
    if len(args) != 2:
        print("ERROR: Invalid parameters. Provide a history ID and new history name.")
        return
    id = find_history(args[0])
    name = args[1]

    gi = connect(context)
    new_history = gi.histories.create_history(name)
    datasets = gi.datasets.get_datasets(history_id=id)
    for dataset in datasets:
        gi.histories.copy_dataset(new_history['id'], dataset['id'])
    print(json.dumps(new_history, indent=4))


def purge(context: Context, args: list):
    if len(args) != 1:
        print("ERROR: Please pass a string used to filter histories to be deleted.")
        print("Use 'abm <cloud> history purge *' to remove ALL histories.")
        return
    all = args[0] == '0'
    gi = connect(context)
    print('Purging histories')
    count = 0
    for history in gi.histories.get_histories():
        if all or args[0] in history['name']:
            gi.histories.delete_history(history['id'], True)
            print(f"Deleted user history {history['id']}")
            count += 1
    for history in gi.histories.get_published_histories():
        if all or args[0] in history['name']:
            gi.histories.delete_history(history['id'], True)
            print(f"Deleted published history {history['id']}")
            count += 1
    print(f'Purged {count} histories')


def tag(context: Context, args: list):
    replace = False
    if '--replace' in args:
        replace = True
        args.remove('--replace')
    if '-r' in args:
        replace = True
        args.remove('-r')
    if len(args) < 2:
        print(
            "ERROR: Invalid command. Please provide the history ID and one or more tags."
        )
        return

    gi = connect(context)
    hid = find_history(gi, args.pop(0))
    if not replace:
        history = gi.histories.show_history(hid, contents=False)
        args += history['tags']
    gi.histories.update_history(hid, tags=args)
    print(f"Set history tags to: {', '.join(args)}")


def summarize(context: Context, args: list):
    if len(args) == 0:
        print("ERROR: Provide one or more history ID values.")
        return
    gi = connect(context)
    all_jobs = []
    while len(args) > 0:
        hid = find_history(gi, args.pop(0))
        history = gi.histories.show_history(history_id=hid)
        jobs = gi.jobs.get_jobs(history_id=hid)
        for job in jobs:
            job['invocation_id'] = ''
            job['history_id'] = hid
            job['history_name'] = history['name']
            job['workflow_id'] = ''
            # if 'workflow_id' in invocation:
            #     job['workflow_id'] = invocation['workflow_id']
            all_jobs.append(job)
        # invocations = gi.invocations.get_invocations(history_id=hid)
        # for invocation in invocations:
        #     id = invocation['id']
        #     #jobs = gi.jobs.get_jobs(history_id=hid, invocation_id=id)
        #     jobs = gi.jobs.get_jobs(history_id=hid)
        #     for job in jobs:
        #         job['invocation_id'] = id
        #         job['history_id'] = hid
        #         if 'workflow_id' in invocation:
        #             job['workflow_id'] = invocation['workflow_id']
        #         all_jobs.append(job)
    # summarize_metrics(gi, gi.jobs.get_jobs(history_id=args[0]))
    summarize_metrics(gi, all_jobs)


def wait(context: Context, args: list):
    state = ''
    if len(args) == 0:
        print("ERROR: No history ID provided")
        return

    gi = connect(context)
    history_id = find_history(gi, args[0])
    if history_id is None:
        print("ERROR: No such history")
        return
    wait_for(gi, history_id)


def wait_for(gi: GalaxyInstance, history_id: str):
    errored = []
    waiting = True
    job_states = JobStates()
    while waiting:
        restart = []
        status_counts = dict()
        terminal = 0
        job_list = gi.jobs.get_jobs(history_id=history_id)
        for job in job_list:
            job_states.update(job)
            state = job['state']
            id = job['id']
            # Count how many jobs are in each state.
            if state not in status_counts:
                status_counts[state] = 1
            else:
                status_counts[state] += 1
            # Count jobs in a terminal state and mark failed jobs for a restart
            if state == 'ok':
                terminal += 1
            elif state == 'error':
                terminal += 1
                if id not in errored:
                    restart.append(id)
                    errored.append(id)
        if len(restart) > 0:
            for job in restart:
                print(f"Restaring job {job}")
                try:
                    gi.jobs.rerun_job(job, remap=True)
                except:
                    try:
                        gi.jobs.rerun_job(job, remap=False)
                    except:
                        print(f"Failed to restart job {job}")
                        waiting = False
        elif len(job_list) == terminal:
            print("All jobs are in a terminal state")
            waiting = False
        if waiting:
            time.sleep(30)
            # elif state == 'paused':
            #     paused += 1
            # print(f"{job['id']}\t{job['state']}\t{job['update_time']}\t{job['tool_id']}")


class JobStates:
    def __init__(self):
        self._jobs = dict()

    def update(self, job):
        id = job['id']
        state = job['state']
        tool = job['tool_id']
        if '/' in tool:
            tool = tool.split('/')[-2]
        if id not in self._jobs:
            print(f"Job {id} {tool} state {state}")
        elif state != self._jobs[id]:
            print(f"Job {id} {tool} {self._jobs[id]} -> {state}")
        self._jobs[id] = state
